#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020, Intel Corporation. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
#

import threading
import itertools
import time
import sys
import concurrent.futures
import os
import fnmatch
from typing import Any
import logging


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

class AsyncError(Exception):
    def __init__(self, error_list, completed_list=None):
        self.error_list = error_list
        self.completed_list = completed_list

    def __str__(self):
        error_list = '\n'.join(self.error_list)
        if self.completed_list:
            completed_list = '\n'.join(['{}:{}'.format(x, y) for x, y in self.completed_list])
            return "Threads reported following errors:\n{}\nFollowing threads completed:\n{}".format(error_list, completed_list)
        return "Threads reported following errors:\n{}".format(error_list)


class DeviceError(Exception):
    device: Any # KMBUSBdevice or KMBPCIeDevice
    message: str

    def __init__(self, device: Any, message: str):
        self.device = device
        self.message = message
    
    def __str__(self):
        return "Device {} failed with error: {}".format(self.device, self.message)


class Signal:
    go = True
    playing = True
    pause = False


def spin(msg, signal):
    write, flush = sys.stdout.write, sys.stdout.flush
    for char in itertools.cycle("|/-\\"):
        if signal.playing:
            status = char + " " + msg
            write(status)
            flush()
            write("\x08" * len(status))
            time.sleep(0.1)
            if signal.pause:
                write(" " * len(status) + "\x08" * len(status))
                signal.playing = False
        if not signal.go:
            break
    write(" " * len(status) + "\x08" * len(status))


from multiprocessing import Pool, TimeoutError, cpu_count, Manager
import os
import itertools
import time

def worker(*args):
    device, func_name, args = args
    try:
        if len(args) > 0:
            result = getattr(device, func_name)(*args)
        else:
            result = getattr(device, func_name)()
        return (device, result)
    except Exception as exc:
        error = '{} generated an exception: {}'.format(device, exc)
        return (device, exc)

def async_dev_ops(devices, func_name, *args, spinner_msg=" Downloading..", max_threads=False):
    error_list = []
    results = {}
    MAX_THREADS = cpu_count()  # Defaults to number of CPUs, adjust if necessary

    if len(devices) > MAX_THREADS:
        raise AsyncError('Too many devices, reduce number of devices to continue.', list(devices))

    iters = [iter(devices)] * MAX_THREADS
    device_groups = itertools.zip_longest(*iters, fillvalue=None)

    for device_group in device_groups:
        pool = Pool(processes=MAX_THREADS)
        signal = Signal()
        spinner = threading.Thread(target=spin, args=(spinner_msg, signal))
        spinner.start()
        worker_args = [(device, func_name, args) for device in device_group if device is not None]
        tasks = {worker_args[0]: pool.apply_async(worker, worker_args) for worker_args in worker_args}
        while True:
            if all([result.ready() for result in tasks.values()]):
                break
            time.sleep(1)
            for device, result in tasks.items():
                
                if result.ready():
                    _, result_or_error = result.get(timeout=1.0)
                    if device in results:
                        continue
                    if isinstance(result_or_error, Exception):
                        logger.warning(f"Device {device} raised an error during {func_name}: {result_or_error}")
                        error_list.append(DeviceError(device, result_or_error))
                    else:
                        logger.debug(f"Device {device} completed {func_name} successfully")
                        results[device] = result_or_error
        pool.close()
        pool.join()
        signal.go = False
        spinner.join()

    if len(error_list) > 0:
        if func_name == 'flash_fip':
            logger.warning(f'Issue flashing FIP: {error_list}')
        else:
            raise AsyncError(error_list, list(results.items()))
    return results


def find_dir(pattern, path):
    for root, dirs, files in os.walk(path):
        for name in dirs:
            if fnmatch.fnmatch(name, pattern):
                return os.path.join(root, name)
    return None

def find_any(pattern, path):
    for root, dirs, files in os.walk(path):
        for name in dirs:
            if fnmatch.fnmatch(name, pattern):
                return os.path.join(root, name)
        for name in files:
            if fnmatch.fnmatch(name, pattern):
                return os.path.join(root, name)
    return None
